"""Ferda related context processors."""
from django.conf import settings
from django_fido.backends import is_fido_backend_used

import ferda
from ferda.plugins import CONFIG
from ferda.settings import SETTINGS


def ferda_modules(request):
    """Return context variables with ferda modules config."""
    if request.resolver_match.app_names:
        current_app = request.resolver_match.app_names[0]   # app_names[0] is app_name from urls.py
    else:
        current_app = None

    ferda_plugins = {}
    for app_name, config in CONFIG.items():
        if config.access(request.user):
            config_dict = config.dict()
            config_dict.pop('access')
            config_dict.pop('permissions')
            ferda_plugins[app_name] = config_dict

    context = {
        'ferda_plugins': ferda_plugins,
        'current_plugin': current_app,
        'current_plugin_name': CONFIG[current_app].name if current_app in CONFIG else None,
    }
    return context


def user_profile(request):
    """Return context variable with user account details."""
    if request.user.is_authenticated:
        context = {
            'user_profile': {
                'username': request.user.username,
                'first_name': request.user.first_name,
                'last_name': request.user.last_name,
                'email': request.user.email,
                'is_staff': request.user.is_staff,
            }
        }
    else:
        context = {}
    return context


def user_permissions(request):
    """Return context variable with user permissions."""
    if request.user.is_authenticated:
        context = {
            'user_permissions': list(request.user.get_all_permissions()),
        }
    else:
        context = {}
    return context


def django_settings(request):
    """Return context variable with selected django settings."""
    context = {
        'django_settings': {
            'TIME_ZONE': settings.TIME_ZONE,
            'LANGUAGE_CODE': settings.LANGUAGE_CODE,
            'LANGUAGES': settings.LANGUAGES,
            'FIDO2_AUTHENTICATION': is_fido_backend_used(),
            'MANUAL_IN_ZONE_DEFAULT_DURATION': SETTINGS.manual_in_zone_default_duration,
            'OBJECT_DETAIL_LOGGER_SERVICES': SETTINGS.object_detail_logger_services,
        }
    }
    return context


def ferda_version(request):
    """Return context variable with Ferda version."""
    context = {
        'ferda_version': {
            'version': ferda.__version__,
        }
    }
    return context
