from datetime import datetime
from typing import Callable, Sequence, Type
from unittest.mock import call
from uuid import UUID

import pytz
from django.test import SimpleTestCase
from fred_api.registry.common_types_pb2 import EmailAddress, Uuid
from fred_api.registry.contact.contact_common_types_pb2 import ContactHistoryId, ContactId
from fred_api.registry.contact.contact_history_types_pb2 import (ContactHistoryRecord, ContactHistoryReply,
                                                                 ContactHistoryRequest)
from fred_api.registry.contact.contact_info_types_pb2 import ContactInfoReply, ContactInfoRequest
from fred_api.registry.contact.contact_state_history_types_pb2 import (ContactStateHistoryReply,
                                                                       ContactStateHistoryRequest)
from fred_api.registry.contact.contact_state_types_pb2 import ContactStateReply, ContactStateRequest
from fred_api.registry.contact.service_search_contact_grpc_pb2 import (SearchContactHistoryReply,
                                                                       SearchContactHistoryRequest, SearchContactReply,
                                                                       SearchContactRequest)
from fred_api.registry.domain.domain_common_types_pb2 import DomainHistoryId, DomainId
from fred_api.registry.domain.domain_history_types_pb2 import DomainHistoryReply, DomainHistoryRequest
from fred_api.registry.domain.domain_info_types_pb2 import DomainInfoReply, DomainInfoRequest
from fred_api.registry.domain.domain_state_history_types_pb2 import DomainStateHistoryReply, DomainStateHistoryRequest
from fred_api.registry.domain.domain_state_types_pb2 import DomainStateReply, DomainStateRequest
from fred_api.registry.domain.service_search_domain_grpc_pb2 import (SearchDomainHistoryReply,
                                                                     SearchDomainHistoryRequest, SearchDomainReply,
                                                                     SearchDomainRequest)
from fred_api.registry.keyset.keyset_common_types_pb2 import KeysetHistoryId, KeysetId
from fred_api.registry.keyset.keyset_history_types_pb2 import KeysetHistoryReply, KeysetHistoryRequest
from fred_api.registry.keyset.keyset_info_types_pb2 import DnsKey, KeysetInfoReply, KeysetInfoRequest
from fred_api.registry.keyset.keyset_state_history_types_pb2 import KeysetStateHistoryReply, KeysetStateHistoryRequest
from fred_api.registry.keyset.keyset_state_types_pb2 import KeysetStateReply, KeysetStateRequest
from fred_api.registry.keyset.service_search_keyset_grpc_pb2 import (SearchKeysetHistoryReply,
                                                                     SearchKeysetHistoryRequest, SearchKeysetReply,
                                                                     SearchKeysetRequest)
from fred_api.registry.nsset.nsset_common_types_pb2 import NssetHistoryId, NssetId
from fred_api.registry.nsset.nsset_history_types_pb2 import NssetHistoryReply, NssetHistoryRequest
from fred_api.registry.nsset.nsset_info_types_pb2 import DnsHost, NssetInfoReply, NssetInfoRequest
from fred_api.registry.nsset.nsset_state_history_types_pb2 import NssetStateHistoryReply, NssetStateHistoryRequest
from fred_api.registry.nsset.nsset_state_types_pb2 import NssetStateReply, NssetStateRequest
from fred_api.registry.nsset.service_search_nsset_grpc_pb2 import (SearchNssetHistoryReply, SearchNssetHistoryRequest,
                                                                   SearchNssetReply, SearchNssetRequest)
from frgal.utils import TestClientMixin
from google.protobuf.message import Message

from ferda.backend.client import FerdaBackendClient
from ferda.backend.decoder import FerdaBackendDecoder


class TestFerdaBackendClient(TestClientMixin, FerdaBackendClient):
    """Test FerdaBackendClient."""


class FerdaBackendClientContactTest(SimpleTestCase):
    """Test FerdaBackendClient's contact methods."""

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    def test_get_contact_info(self):
        contact_info = ContactInfoReply()
        contact_info.data.name = 'Harry Potter'
        contact_info.data.emails.extend([
            EmailAddress(value='harry@example.com'),
            EmailAddress(value='potter@example.com'),
        ])
        contact_info.data.publish['name'] = True
        contact_info.data.publish['emails'] = True
        contact_info.data.publish['notify_emails'] = False
        self.client.mock.return_value = contact_info

        response = self.client.get_contact_info(UUID(int=42))
        self.assertEqual(response['name'], 'Harry Potter')
        self.assertEqual(response['email'], 'harry@example.com, potter@example.com')
        self.assertEqual(response['notify_email'], None)
        self.assertEqual(response['publish'], {'email': True, 'name': True, 'notify_email': False})

        response = self.client.get_contact_info(UUID(int=42), history_id=UUID(int=42))
        contact_id = ContactId(uuid=Uuid(value=str(UUID(int=42))))
        history_id = ContactHistoryId(uuid=Uuid(value=str(UUID(int=42))))
        self.client.mock.assert_has_calls([
            call(
                ContactInfoRequest(contact_id=contact_id),
                method='/Fred.Registry.Api.Contact.Contact/get_contact_info',
                timeout=None,
            ),
            call(
                ContactInfoRequest(contact_id=contact_id, contact_history_id=history_id),
                method='/Fred.Registry.Api.Contact.Contact/get_contact_info',
                timeout=None,
            ),
        ])

    def test_get_contact_history(self):
        reply = ContactHistoryReply()
        reply.data.contact_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_contact_history(contact_id=UUID(int=42)), {
            'contact_id': UUID(int=42),
            'timeline': [],
            'valid_to': None,
        })

        request = ContactHistoryRequest()
        request.contact_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_has_calls([
            call(
                request,
                method='/Fred.Registry.Api.Contact.Contact/get_contact_history',
                timeout=None,
            ),
        ])

    def test_get_contact_history_id_by_datetime(self):
        contact_history_record = ContactHistoryRecord()
        contact_history_record.contact_history_id.uuid.value = str(UUID(int=43))
        contact_history_record.valid_from.FromDatetime(datetime(2019, 1, 1))

        reply = ContactHistoryReply()
        reply.data.contact_id.uuid.value = str(UUID(int=42))
        reply.data.timeline.extend([contact_history_record])
        reply.data.valid_to.FromDatetime(datetime(2020, 1, 1))

        self.client.mock.return_value = reply

        response = self.client.get_contact_history_id_by_datetime(UUID(int=42), datetime(2019, 5, 5))
        self.assertEqual(response, UUID(int=43))

        request = ContactHistoryRequest()
        request.contact_id.uuid.value = str(UUID(int=42))
        request.history.lower_limit.timestamp.FromDatetime(datetime(2019, 5, 5))
        request.history.upper_limit.timestamp.FromDatetime(datetime(2019, 5, 5))
        self.client.mock.assert_has_calls([
            call(
                request,
                method='/Fred.Registry.Api.Contact.Contact/get_contact_history',
                timeout=None,
            ),
        ])

    def test_get_contact_state(self):
        reply = ContactStateReply()
        reply.data.contact_id.uuid.value = str(UUID(int=42))
        reply.data.state.flags['linked'] = True
        reply.data.state.flags['validated'] = False
        self.client.mock.return_value = reply

        self.assertEqual(
            self.client.get_contact_state(UUID(int=42)),
            {'linked': True, 'validated': False}
        )

        request = ContactStateRequest()
        request.contact_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Contact.Contact/get_contact_state',
            timeout=None,
        )

    def test_get_contact_state_history(self):
        reply = ContactStateHistoryReply()
        reply.data.contact_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_contact_state_history(contact_id=UUID(int=42)), {
            'contact_id': UUID(int=42),
            'history': {'timeline': [], 'valid_to': None},
        })

        request = ContactStateHistoryRequest()
        request.contact_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Contact.Contact/get_contact_state_history',
            timeout=None,
        )


class FerdaBackendClientDomainTest(SimpleTestCase):
    """Test FerdaBackendClient's domain methods."""

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    def test_get_domain_info(self):
        reply = DomainInfoReply()
        reply.data.fqdn.value = 'example.com'
        self.client.mock.return_value = reply

        response = self.client.get_domain_info(UUID(int=42))
        self.assertEqual(response['fqdn'], 'example.com')

        self.client.get_domain_info(UUID(int=42), history_id=UUID(int=43))

        domain_id = DomainId()
        domain_id.uuid.value = str(UUID(int=42))
        history_id = DomainHistoryId()
        history_id.uuid.value = str(UUID(int=43))
        self.client.mock.assert_has_calls([
            call(
                DomainInfoRequest(domain_id=domain_id),
                method='/Fred.Registry.Api.Domain.Domain/get_domain_info',
                timeout=None,
            ),
            call(
                DomainInfoRequest(domain_id=domain_id, domain_history_id=history_id),
                method='/Fred.Registry.Api.Domain.Domain/get_domain_info',
                timeout=None,
            ),
        ])

    def test_get_domain_history(self):
        reply = DomainHistoryReply()
        reply.data.domain_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_domain_history(domain_id=UUID(int=42)), {
            'domain_id': UUID(int=42),
            'timeline': [],
            'valid_to': None,
        })

        request = DomainHistoryRequest()
        request.domain_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_has_calls([
            call(
                request,
                method='/Fred.Registry.Api.Domain.Domain/get_domain_history',
                timeout=None,
            ),
        ])

    def test_get_domain_state(self):
        reply = DomainStateReply()
        reply.data.domain_id.uuid.value = str(UUID(int=42))
        reply.data.state.flags['linked'] = True
        reply.data.state.flags['serverBlocked'] = False
        self.client.mock.return_value = reply

        self.assertEqual(
            self.client.get_domain_state(UUID(int=42)),
            {'linked': True, 'serverBlocked': False}
        )

        request = DomainStateRequest()
        request.domain_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Domain.Domain/get_domain_state',
            timeout=None,
        )

    def test_get_domain_state_history(self):
        reply = DomainStateHistoryReply()
        reply.data.domain_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_domain_state_history(domain_id=UUID(int=42)), {
            'domain_id': UUID(int=42),
            'history': {'timeline': [], 'valid_to': None},
        })

        request = DomainStateHistoryRequest()
        request.domain_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Domain.Domain/get_domain_state_history',
            timeout=None,
        )


class FerdaBackendClientKeysetTest(SimpleTestCase):
    """Test FerdaBackendClient's keyset methods."""

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    def test_get_keyset_info(self):
        dns_key = DnsKey()
        dns_key.flags = 257
        dns_key.protocol = 3
        dns_key.alg = 13
        dns_key.key = 'whatever=='
        reply = KeysetInfoReply()
        reply.data.keyset_id.uuid.value = str(UUID(int=42))
        reply.data.dns_keys.extend([dns_key])
        self.client.mock.return_value = reply

        response = self.client.get_keyset_info(UUID(int=42))
        self.assertEqual(response['keyset_id'], UUID(int=42))
        self.assertEqual(response['dns_keys'], [{'flags': 257, 'protocol': 3, 'alg': 13, 'key': 'whatever=='}])

        self.client.get_keyset_info(UUID(int=42), history_id=UUID(int=43))

        keyset_id = KeysetId()
        keyset_id.uuid.value = str(UUID(int=42))
        history_id = KeysetHistoryId()
        history_id.uuid.value = str(UUID(int=43))
        self.client.mock.assert_has_calls([
            call(
                KeysetInfoRequest(keyset_id=keyset_id),
                method='/Fred.Registry.Api.Keyset.Keyset/get_keyset_info',
                timeout=None,
            ),
            call(
                KeysetInfoRequest(keyset_id=keyset_id, keyset_history_id=history_id),
                method='/Fred.Registry.Api.Keyset.Keyset/get_keyset_info',
                timeout=None,
            ),
        ])

    def test_get_keyset_history(self):
        reply = KeysetHistoryReply()
        reply.data.keyset_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_keyset_history(keyset_id=UUID(int=42)), {
            'keyset_id': UUID(int=42),
            'timeline': [],
            'valid_to': None,
        })

        request = KeysetHistoryRequest()
        request.keyset_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_has_calls([
            call(
                request,
                method='/Fred.Registry.Api.Keyset.Keyset/get_keyset_history',
                timeout=None,
            ),
        ])

    def test_get_keyset_state(self):
        reply = KeysetStateReply()
        reply.data.keyset_id.uuid.value = str(UUID(int=42))
        reply.data.state.flags['linked'] = True
        reply.data.state.flags['serverBlocked'] = False
        self.client.mock.return_value = reply

        self.assertEqual(
            self.client.get_keyset_state(UUID(int=42)),
            {'linked': True, 'serverBlocked': False}
        )

        request = KeysetStateRequest()
        request.keyset_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Keyset.Keyset/get_keyset_state',
            timeout=None,
        )

    def test_get_keyset_state_history(self):
        reply = KeysetStateHistoryReply()
        reply.data.keyset_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_keyset_state_history(keyset_id=UUID(int=42)), {
            'keyset_id': UUID(int=42),
            'history': {'timeline': [], 'valid_to': None},
        })

        request = KeysetStateHistoryRequest()
        request.keyset_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Keyset.Keyset/get_keyset_state_history',
            timeout=None,
        )


class FerdaBackendClientNssetTest(SimpleTestCase):
    """Test FerdaBackendClient's nsset methods."""

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    def test_get_nsset_info(self):
        dns_host = DnsHost()
        dns_host.fqdn = 'dns.example.com'
        dns_host.ip_addresses[:] = ['127.0.0.1']
        reply = NssetInfoReply()
        reply.data.nsset_id.uuid.value = str(UUID(int=42))
        reply.data.dns_hosts.extend([dns_host])
        self.client.mock.return_value = reply

        response = self.client.get_nsset_info(UUID(int=42))
        self.assertEqual(response['nsset_id'], UUID(int=42))
        self.assertEqual(response['dns_hosts'], [{'fqdn': 'dns.example.com', 'ip_addresses': ['127.0.0.1']}])

        self.client.get_nsset_info(UUID(int=42), history_id=UUID(int=43))

        nsset_id = NssetId()
        nsset_id.uuid.value = str(UUID(int=42))
        history_id = NssetHistoryId()
        history_id.uuid.value = str(UUID(int=43))
        self.client.mock.assert_has_calls([
            call(
                NssetInfoRequest(nsset_id=nsset_id),
                method='/Fred.Registry.Api.Nsset.Nsset/get_nsset_info',
                timeout=None,
            ),
            call(
                NssetInfoRequest(nsset_id=nsset_id, nsset_history_id=history_id),
                method='/Fred.Registry.Api.Nsset.Nsset/get_nsset_info',
                timeout=None,
            ),
        ])

    def test_get_nsset_history(self):
        reply = NssetHistoryReply()
        reply.data.nsset_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_nsset_history(nsset_id=UUID(int=42)), {
            'nsset_id': UUID(int=42),
            'timeline': [],
            'valid_to': None,
        })

        request = NssetHistoryRequest()
        request.nsset_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_has_calls([
            call(
                request,
                method='/Fred.Registry.Api.Nsset.Nsset/get_nsset_history',
                timeout=None,
            ),
        ])

    def test_get_nsset_state(self):
        reply = NssetStateReply()
        reply.data.nsset_id.uuid.value = str(UUID(int=42))
        reply.data.state.flags['linked'] = True
        reply.data.state.flags['serverBlocked'] = False
        self.client.mock.return_value = reply

        self.assertEqual(
            self.client.get_nsset_state(UUID(int=42)),
            {'linked': True, 'serverBlocked': False}
        )

        request = NssetStateRequest()
        request.nsset_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Nsset.Nsset/get_nsset_state',
            timeout=None,
        )

    def test_get_nsset_state_history(self):
        reply = NssetStateHistoryReply()
        reply.data.nsset_id.uuid.value = str(UUID(int=42))
        self.client.mock.return_value = reply

        self.assertEqual(self.client.get_nsset_state_history(nsset_id=UUID(int=42)), {
            'nsset_id': UUID(int=42),
            'history': {'timeline': [], 'valid_to': None},
        })

        request = NssetStateHistoryRequest()
        request.nsset_id.uuid.value = str(UUID(int=42))
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Nsset.Nsset/get_nsset_state_history',
            timeout=None,
        )


class SearchObjectTestMixin:
    """Test FerdaBackendClient's search methods."""

    request_type: Type[Message]
    reply_type: Type[Message]
    search_method_name: str
    grpc_method_name: str
    query_values: Sequence[str]
    searched_items: Sequence[str]

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    @property
    def search_method(self) -> Callable:
        return getattr(self.client, self.search_method_name)

    def test_search_object_empty(self: SimpleTestCase):
        reply = self.reply_type()
        reply.data.results.extend([])
        self.client.mock.return_value = reply

        self.assertEqual(self.search_method(query_values=[]), {
            'results': [],
            'estimated_total': {
                'lower_estimate': 0,
                'upper_estimate': 0,
            },
            'searched_items': [],
        })

        request = self.request_type()
        request.query_values[:] = []
        self.client.mock.assert_called_once_with(
            request,
            method=self.grpc_method_name,
            timeout=None,
        )

    def test_search_object_not_empty(self: SimpleTestCase):
        result = self.reply_type.Data.Result()  # type: ignore
        result.object_id.uuid.value = str(UUID(int=42))
        result.matched_items[:] = self.searched_items

        reply = self.reply_type()
        reply.data.results.extend([result])
        reply.data.searched_items[:] = self.searched_items
        self.client.mock.return_value = reply

        self.assertEqual(
            self.search_method(query_values=self.query_values, limit=10, searched_items=self.searched_items), {
                'results': [{
                    'object_id': UUID(int=42),
                    'matched_items': list(self.searched_items),
                }],
                'estimated_total': {
                    'lower_estimate': 0,
                    'upper_estimate': 0,
                },
                'searched_items': list(self.searched_items),
            }
        )

        request = self.request_type()
        request.query_values[:] = self.query_values
        request.limit = 10
        request.searched_items[:] = self.searched_items
        self.client.mock.assert_called_once_with(
            request,
            method=self.grpc_method_name,
            timeout=None,
        )


class SearchObjectHistoryTestMixin:
    """Test FerdaBackendClient's search history methods."""

    request_type: Type[Message]
    reply_type: Type[Message]
    object_history_id_type: Type[Message]
    object_history_ids_name: str
    max_number_of_objects_name: str
    search_method_name: str
    grpc_method_name: str
    query_values: Sequence[str]
    searched_items: Sequence[str]

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    @property
    def search_method(self) -> Callable:
        return getattr(self.client, self.search_method_name)

    def test_search_object_history_empty(self: SimpleTestCase):
        reply = self.reply_type()
        reply.data.results.extend([])
        self.client.mock.return_value = reply

        self.assertEqual(self.search_method(query_values=[]), {
            'results': [],
            'estimated_total': {
                'lower_estimate': 0,
                'upper_estimate': 0,
            },
            'searched_items': [],
        })

        request = self.request_type()
        request.query_values[:] = []
        self.client.mock.assert_called_once_with(
            request,
            method=self.grpc_method_name,
            timeout=None,
        )

    def test_search_object_history_not_empty(self: SimpleTestCase):
        history_period = self.reply_type.Data.Result.HistoryPeriod()  # type: ignore
        history_period.matched_items[:] = self.searched_items
        history_period.valid_from.FromDatetime(datetime(2019, 1, 1, tzinfo=pytz.utc))
        getattr(history_period, self.object_history_ids_name).extend([
            self.object_history_id_type(uuid=Uuid(value=str(UUID(int=43)))),
            self.object_history_id_type(uuid=Uuid(value=str(UUID(int=44)))),
        ])
        search_result = self.reply_type.Data.Result()  # type: ignore
        search_result.object_id.uuid.value = str(UUID(int=42))
        search_result.histories.extend([history_period])
        reply = self.reply_type()
        reply.data.results.extend([search_result])
        reply.data.searched_items[:] = self.searched_items
        self.client.mock.return_value = reply

        self.assertEqual(
            self.search_method(query_values=self.query_values, limit=10, searched_items=self.searched_items), {
                'results': [{
                    'object_id': UUID(int=42),
                    'histories': [{
                        'valid_from': datetime(2019, 1, 1, tzinfo=pytz.utc),
                        'valid_to': None,
                        'matched_items': list(self.searched_items),
                        self.object_history_ids_name: [UUID(int=43), UUID(int=44)],
                    }],
                }],
                'estimated_total': {
                    'lower_estimate': 0,
                    'upper_estimate': 0,
                },
                'searched_items': list(self.searched_items),
            }
        )

        request = self.request_type()
        request.query_values[:] = self.query_values
        setattr(request, self.max_number_of_objects_name, 10)
        request.searched_items[:] = self.searched_items
        self.client.mock.assert_called_once_with(
            request,
            method=self.grpc_method_name,
            timeout=None,
        )


class FerdaBackendClientSearchContactTest(SearchObjectTestMixin, SimpleTestCase):
    request_type = SearchContactRequest
    reply_type = SearchContactReply
    search_method_name = 'search_contact'
    grpc_method_name = '/Fred.Registry.Api.Contact.SearchContact/search_contact'
    query_values = ['HARRY']
    searched_items = ('contact_handle',)


class FerdaBackendClientSearchDomainTest(SearchObjectTestMixin, SimpleTestCase):
    request_type = SearchDomainRequest
    reply_type = SearchDomainReply
    search_method_name = 'search_domain'
    grpc_method_name = '/Fred.Registry.Api.Domain.SearchDomain/search_domain'
    query_values = ['example.com']
    searched_items = ('fqdn',)


class FerdaBackendClientSearchNssetTest(SearchObjectTestMixin, SimpleTestCase):
    request_type = SearchNssetRequest
    reply_type = SearchNssetReply
    search_method_name = 'search_nsset'
    grpc_method_name = '/Fred.Registry.Api.Nsset.SearchNsset/search_nsset'
    query_values = ['NSSET']
    searched_items = ('nsset_handle',)


class FerdaBackendClientSearchKeysetTest(SearchObjectTestMixin, SimpleTestCase):
    request_type = SearchKeysetRequest
    reply_type = SearchKeysetReply
    search_method_name = 'search_keyset'
    grpc_method_name = '/Fred.Registry.Api.Keyset.SearchKeyset/search_keyset'
    query_values = ['KEYSET']
    searched_items = ('keyset_handle',)


class FerdaBackendClientSearchContactHistoryTest(SearchObjectHistoryTestMixin, SimpleTestCase):
    request_type = SearchContactHistoryRequest
    reply_type = SearchContactHistoryReply
    object_history_id_type = ContactHistoryId
    object_history_ids_name = 'contact_history_ids'
    max_number_of_objects_name = 'max_number_of_contacts'
    search_method_name = 'search_contact_history'
    grpc_method_name = '/Fred.Registry.Api.Contact.SearchContact/search_contact_history'
    query_values = ['HARRY']
    searched_items = ('contact_handle',)


class FerdaBackendClientSearchDomainHistoryTest(SearchObjectHistoryTestMixin, SimpleTestCase):
    request_type = SearchDomainHistoryRequest
    reply_type = SearchDomainHistoryReply
    object_history_id_type = DomainHistoryId
    object_history_ids_name = 'domain_history_ids'
    max_number_of_objects_name = 'max_number_of_domains'
    search_method_name = 'search_domain_history'
    grpc_method_name = '/Fred.Registry.Api.Domain.SearchDomain/search_domain_history'
    query_values = ['example.com']
    searched_items = ('fqdn',)


class FerdaBackendClientSearchNssetHistoryTest(SearchObjectHistoryTestMixin, SimpleTestCase):
    request_type = SearchNssetHistoryRequest
    reply_type = SearchNssetHistoryReply
    object_history_id_type = NssetHistoryId
    object_history_ids_name = 'nsset_history_ids'
    max_number_of_objects_name = 'max_number_of_nssets'
    search_method_name = 'search_nsset_history'
    grpc_method_name = '/Fred.Registry.Api.Nsset.SearchNsset/search_nsset_history'
    query_values = ['NSSET']
    searched_items = ('nsset_handle',)


class FerdaBackendClientSearchKeysetHistoryTest(SearchObjectHistoryTestMixin, SimpleTestCase):
    request_type = SearchKeysetHistoryRequest
    reply_type = SearchKeysetHistoryReply
    object_history_id_type = KeysetHistoryId
    object_history_ids_name = 'keyset_history_ids'
    max_number_of_objects_name = 'max_number_of_keysets'
    search_method_name = 'search_keyset_history'
    grpc_method_name = '/Fred.Registry.Api.Keyset.SearchKeyset/search_keyset_history'
    query_values = ['KEYSET']
    searched_items = ('keyset_handle',)


class FerdaBackendClientCommonTest(SimpleTestCase):
    """Test FerdaBackendClient's common methods."""

    def setUp(self):
        self.client = TestFerdaBackendClient(decoder=FerdaBackendDecoder())

    def test_set_history_limits_by_history_id(self):
        request = ContactHistoryRequest()
        result = self.client._set_history_limits(request, start=UUID(int=43), end=UUID(int=44))
        self.assertEqual(result.history.lower_limit.contact_history_id.uuid.value, str(UUID(int=43)))
        self.assertEqual(result.history.upper_limit.contact_history_id.uuid.value, str(UUID(int=44)))

    def test_set_history_limits_by_timestamp(self):
        request = ContactHistoryRequest()
        result = self.client._set_history_limits(
            request,
            start=datetime.fromtimestamp(0, tz=pytz.utc),
            end=datetime.fromtimestamp(44442222.1199, tz=pytz.utc),
        )
        self.assertEqual(result.history.lower_limit.timestamp.seconds, 0)
        self.assertEqual(result.history.upper_limit.timestamp.seconds, 44442222)
        self.assertEqual(result.history.upper_limit.timestamp.nanos, 119900000)
