from datetime import datetime
from unittest.mock import patch
from uuid import UUID

import pytz
from django.test import TestCase, override_settings
from django.urls import reverse
from django_countries.fields import Country
from freezegun import freeze_time
from grill import LogProperty

from ferda.backend.exceptions import ContactDoesNotExist
from ferda.constants import LogEntryType, LoggerResult
from ferda.models import ContactRepresentative, User
from ferda.registry.plugins.badges import Badge
from ferda.tests.utils import LoggerMixin

from .common import GenericInfoViewTest, GenericStateHistoryViewTest, GenericStateViewTest

CONTACT_REPRESENTATIVE_DATA = {
    'contact_uuid': UUID(int=1),
    'organization': 'Sherlock Holmes',
    'street1': '221B Baker Street',
    'city': 'London',
    'postal_code': 'NW1 6XE',
    'country': Country('GB'),
    'valid_from': datetime(2020, 1, 1, tzinfo=pytz.utc),
}


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactInfoView(GenericInfoViewTest, TestCase):
    object_type = 'contact'
    object_does_not_exist = ContactDoesNotExist
    log_entry_type = LogEntryType.CONTACT_DETAIL


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactStateView(GenericStateViewTest, TestCase):
    object_type = 'contact'
    object_does_not_exist = ContactDoesNotExist
    log_entry_type = LogEntryType.CONTACT_STATE


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactStateHistoryView(GenericStateHistoryViewTest, TestCase):
    object_type = 'contact'
    object_does_not_exist = ContactDoesNotExist
    log_entry_type = LogEntryType.CONTACT_STATE


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactRepresentative(LoggerMixin, TestCase):

    def setUp(self):
        super().setUp()

        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    def test_get(self):
        ContactRepresentative.objects.create(**CONTACT_REPRESENTATIVE_DATA)
        response = self.client.get(
            reverse('registry:contact:ajax_representative', kwargs={'contact_id': UUID(int=1)})
        )
        self.assertJSONEqual(
            response.content, {
                'contact_id': str(UUID(int=1)),
                'company': 'Sherlock Holmes',
                'street': ['221B Baker Street'],
                'city': 'London',
                'postal_code': 'NW1 6XE',
                'state_or_province': '',
                'country': 'United Kingdom',
                'country_code': 'GB',
            },
        )

    def test_does_not_exist(self):
        response = self.client.get(
            reverse('registry:contact:ajax_representative', kwargs={'contact_id': UUID(int=2)})
        )
        self.assertEqual(response.status_code, 404)


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactRepresentativeChangeView(LoggerMixin, TestCase):

    def setUp(self):
        super().setUp()
        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    def test_add(self):
        self.check()
        self.assertQuerysetEqual(
            ContactRepresentative.objects.values_list('contact_uuid', 'street1', 'street2', 'city', 'country',
                                                      'valid_from', 'valid_to'),
            [
                (UUID(int=1), 'Test lane 42', '', 'Testville', 'CZ',
                 datetime(2020, 2, 2, tzinfo=pytz.utc), None),
            ],
            transform=tuple,
        )

    def test_change(self):
        ContactRepresentative.objects.create(**CONTACT_REPRESENTATIVE_DATA)
        self.check()
        self.assertQuerysetEqual(
            ContactRepresentative.objects.values_list('contact_uuid', 'street1', 'street2', 'city', 'country',
                                                      'valid_from', 'valid_to'),
            [
                (UUID(int=1), '221B Baker Street', '', 'London', 'GB',
                 datetime(2020, 1, 1, tzinfo=pytz.utc), datetime(2020, 2, 2, tzinfo=pytz.utc)),
                (UUID(int=1), 'Test lane 42', '', 'Testville', 'CZ',
                 datetime(2020, 2, 2, tzinfo=pytz.utc), None),
            ],
            transform=tuple,
            ordered=False,
        )

    def test_no_change(self):
        ContactRepresentative.objects.create(**CONTACT_REPRESENTATIVE_DATA)
        response = self.client.post(
            reverse('registry:contact:ajax_representative_change', kwargs={'contact_id': UUID(int=1)}),
            data=CONTACT_REPRESENTATIVE_DATA,
        )
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(
            ContactRepresentative.objects.values_list('contact_uuid', 'street1', 'country', 'valid_from', 'valid_to'),
            [(UUID(int=1), '221B Baker Street', 'GB', datetime(2020, 1, 1, tzinfo=pytz.utc), None)],
            transform=tuple,
        )

    @freeze_time('2020-02-02')
    def check(self):
        response = self.client.post(
            reverse('registry:contact:ajax_representative_change', kwargs={'contact_id': UUID(int=1)}),
            data={
                'organization': 'Test corp.',
                'street1': 'Test lane 42',
                'street2': '',
                'street3': '',
                'city': 'Testville',
                'postal_code': 12345,
                'state_or_province': '',
                'country': 'CZ',
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertLog(
            LogEntryType.CONTACT_REPRESENTATIVE_CHANGE,
            LoggerResult.SUCCESS,
            input_properties={'uuid': LogProperty(str(UUID(int=1)), None)},
            references={'contact': str(UUID(int=1))},
        )

    def test_bad_request(self):
        response = self.client.post(
            reverse('registry:contact:ajax_representative_change', kwargs={'contact_id': UUID(int=1)}),
            data={},
        )
        self.assertEqual(response.status_code, 400)
        self.assertLog(
            LogEntryType.CONTACT_REPRESENTATIVE_CHANGE,
            LoggerResult.FAIL,
            input_properties={'uuid': LogProperty(str(UUID(int=1)), None)},
            references={'contact': str(UUID(int=1))},
        )


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactRepresentativeDeleteView(LoggerMixin, TestCase):

    def setUp(self):
        super().setUp()
        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    @freeze_time('2020-02-02')
    def test_delete(self):
        ContactRepresentative.objects.create(**CONTACT_REPRESENTATIVE_DATA)
        response = self.client.post(
            reverse('registry:contact:ajax_representative_delete', kwargs={'contact_id': UUID(int=1)})
        )
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(
            ContactRepresentative.objects.values_list('contact_uuid', 'country', 'valid_to'),
            [(UUID(int=1), 'GB', datetime(2020, 2, 2, tzinfo=pytz.utc))],
            transform=tuple,
        )
        self.assertLog(
            LogEntryType.CONTACT_REPRESENTATIVE_DELETE,
            LoggerResult.SUCCESS,
            input_properties={'uuid': LogProperty(str(UUID(int=1)), None)},
            references={'contact': str(UUID(int=1))},
        )

    def test_does_not_exist(self):
        response = self.client.post(
            reverse('registry:contact:ajax_representative_delete', kwargs={'contact_id': UUID(int=2)})
        )
        self.assertEqual(response.status_code, 404)
        self.assertLog(
            LogEntryType.CONTACT_REPRESENTATIVE_DELETE,
            LoggerResult.FAIL,
            input_properties={'uuid': LogProperty(str(UUID(int=2)), None)},
            references={'contact': str(UUID(int=2))},
        )


@override_settings(ROOT_URLCONF='ferda.tests.urls')
class TestContactBadges(LoggerMixin, TestCase):

    def setUp(self):
        super().setUp()

        self.user = User.objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client.force_login(self.user)

    @patch('ferda.views.ajax.contact.get_contact_badges', autospec=True)
    def test_get(self, func_mock):
        func_mock.return_value = [Badge(name='prefect', image='http://example.com/badge.png')]
        response = self.client.get(
            reverse('registry:contact:ajax_badges', kwargs={'contact_id': UUID(int=42)})
        )
        self.assertJSONEqual(
            response.content, [
                {'name': 'prefect', 'image': 'http://example.com/badge.png'},
            ]
        )
